from typing import Optional
from typing import Sequence

from typing_extensions import Literal
from typing_extensions import TypedDict
from typing_extensions import final


class Schema(TypedDict, total=False):
    title: str
    description: str
    type: Literal["array", "string", "float", "number"]
    format: str
    examples: Sequence[object]
    minimum: Optional[float]
    maximum: Optional[float]
    exclusiveMinimum: Optional[float]
    exclusiveMaximum: Optional[float]
    minItems: Optional[int]
    maxItems: Optional[int]


class SchemaField:
    @classmethod
    @final
    def __modify_schema__(cls, field_schema: dict) -> None:
        """
        This final method is called by pydantic and collects overrides from
        :func:`Phantom.__schema__() <phantom.Phantom.__schema__>`.
        Override :func:`__schema__() <phantom.Phantom.__schema__>` to provide custom
        schema representations for phantom types.
        """
        field_schema.update(
            {key: value for key, value in cls.__schema__().items() if value is not None}
        )

    @classmethod
    def __schema__(cls) -> Schema:
        """
        Hook for providing schema metadata. Override in subclasses to customize a types
        schema representation. See pydantic's documentation on ``__modify_schema__()``
        for more information. This hook differs to pydantic's ``__modify_schema__()``
        and expects subclasses to instantiate new dicts instead of mutating a given one.

        Example:

        .. code-block:: python

            class Name(str, Phantom, predicate=...):
                @classmethod
                def __schema__(cls):
                    return {**super().__schema__(), "description": "A name type"}
        """
        return {"title": cls.__name__}



"""
Types for narrowing on the builtin datetime types.
"""

import datetime

from .base import Phantom
from .predicates.datetime import is_tz_aware
from .predicates.datetime import is_tz_naive
from .schema import Schema


class TZAware(datetime.datetime, Phantom, predicate=is_tz_aware):
    """
    A type for helping ensure that ``datetime`` objects are always timezone aware.

    >>> isinstance(datetime.datetime.now(), TZAware)
    False
    >>> isinstance(datetime.datetime.now(tz=datetime.timezone.utc), TZAware)
    True
    """

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A date-time with timezone data.",
        }


class TZNaive(datetime.datetime, Phantom, predicate=is_tz_naive):
    """
    >>> isinstance(datetime.datetime.now(), TZNaive)
    True
    >>> isinstance(datetime.datetime.now(tz=datetime.timezone.utc), TZNaive)
    False
    """

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A date-time without timezone data.",
            "format": "date-time-naive",
        }



"""
Use ``Phantom`` to create arbitrary phantom types using boolean predicates.

.. code-block:: python

    import phantom


    def is_big(value: int) -> bool:
        return value > 5


    class Big(int, Phantom, predicate=is_big):
        ...


    assert isinstance(10, Big)  # this passes
"""
from .base import Phantom
from .base import PhantomBase
from .base import PhantomMeta
from .base import get_bound_parser
from .predicates.base import Predicate

__version__ = "0.15.0"
__all__ = ("PhantomBase", "Phantom", "Predicate", "get_bound_parser", "PhantomMeta")



# Workaround for something that looks like this bug
# https://github.com/pytest-dev/pytest/issues/4386
from __future__ import annotations

import types
from dataclasses import is_dataclass
from itertools import product
from typing import MutableMapping
from typing import MutableSequence
from typing import MutableSet
from typing import NewType
from typing import Tuple
from typing import Union

from typing_extensions import Final
from typing_extensions import TypeGuard
from typing_extensions import get_args
from typing_extensions import get_origin


class UnresolvedClassAttribute(NotImplementedError):
    ...


def resolve_class_attr(
    cls: type,
    name: str,
    argument: object | None,
    required: bool = True,
) -> None:
    argument = getattr(cls, name, None) if argument is None else argument
    if argument is not None:
        setattr(cls, name, argument)
    elif required and not getattr(cls, "__abstract__", False):
        raise UnresolvedClassAttribute(
            f"Concrete phantom type {cls.__qualname__} must define class attribute "
            f"{name}."
        )


BoundType = Union[type, Tuple[type, ...]]


def _is_union(type_: BoundType) -> bool:
    return get_origin(type_) is Union


def _is_intersection(type_: BoundType) -> bool:
    return isinstance(type_, tuple)


def is_subtype(a: BoundType, b: BoundType) -> bool:  # noqa: C901
    """
    Return True if ``a`` is a subtype of ``b``. Supports single-level typing.Unions
    and intersections represented as tuples respectively without nesting.

    The cases that this function deals with can be divided into the following cases,
    where T is neither a union or intersection:

    1. Union, Union: Success if all types in a have a subclass in b.
    2. T, Union: Success if a is a subclass of one or more types in b.
    3. Union, T: Always fails (except when a single-type union, see 9).
    4. T, Intersection: Success if a is a subclass of all types in b.
    5. Intersection, T: Success if one type in a is a subclass of b.
    6. Intersection, Union: Success if one type in a is a subclass of one type in b.
    7. Union, Intersection: Always fails (except when a is a single-type union see 4).
    8. Intersection, Intersection: Success if all items in b have a subclass in a.
    9. T, T: Success if a is a subclass of b.
    """

    if _is_union(a) and _is_union(b):
        for a_part in get_args(a):
            for b_part in get_args(b):
                if issubclass(a_part, b_part):
                    break
            else:
                return False
        return True
    elif _is_intersection(a) and _is_union(b):
        assert isinstance(a, tuple)
        for a_part, b_part in product(a, get_args(b)):
            if issubclass(a_part, b_part):
                return True
        return False
    elif _is_intersection(a) and _is_intersection(b):
        assert isinstance(a, tuple)
        assert isinstance(b, tuple)
        for b_part in b:
            for a_part in a:
                if issubclass(a_part, b_part):
                    break
            else:
                return False
        return True
    elif _is_union(a):
        return False
    elif _is_union(b):
        assert isinstance(a, type)
        return any(issubclass(a, b_part) for b_part in get_args(b))
    elif _is_intersection(b):
        assert isinstance(a, type)
        assert isinstance(b, tuple)
        return all(issubclass(a, b_part) for b_part in b)
    elif _is_intersection(a):
        assert isinstance(a, tuple)
        return any(issubclass(a_part, b) for a_part in a)
    assert isinstance(a, type)
    return issubclass(a, b)


def fully_qualified_name(cls: type) -> str:
    return f"{cls.__module__}.{cls.__qualname__}"


mutable: Final = (MutableSequence, MutableSet, MutableMapping)
NotKnownMutable = NewType("NotKnownMutable", type)
"""
Internal type to mark types that are not known to be mutable. The term immutable is
avoided here because there is no way to guarantee that a checked type isn't actually
mutable, so we don't want to communicate in strong terms here.
"""


class MutableType(TypeError):
    ...


def is_not_mutable(type_: BoundType) -> TypeGuard[NotKnownMutable]:
    if any(is_subtype(type_, mutable_type) for mutable_type in mutable):
        raise MutableType(f"{type_!r} is a subclass of one of {mutable!r}")
    if (
        is_dataclass(type_)
        and not type_.__dataclass_params__.frozen  # type: ignore[union-attr]
    ):
        raise MutableType(f"{type_!r} is a an unfrozen dataclass type")
    return True


MaybeUnionType: type | None
try:
    MaybeUnionType = types.UnionType  # type: ignore[attr-defined]
except AttributeError:
    MaybeUnionType = None


def is_union_type(value: object) -> TypeGuard[type]:
    return MaybeUnionType is not None and isinstance(value, MaybeUnionType)


def is_union(value: object) -> TypeGuard[type]:
    return get_origin(value) == Union or is_union_type(value)



"""
Requires the phonenumbers_ package which can be installed with:

.. _phonenumbers: https://pypi.org/project/phonenumbers/

.. code-block:: bash

    $ python3 -m pip install phantom-types[phonenumbers]
"""
from __future__ import annotations

from typing import cast

import phonenumbers
from typing_extensions import Final

from phantom import Phantom
from phantom import get_bound_parser
from phantom.fn import excepts
from phantom.schema import Schema

__all__ = (
    "InvalidPhoneNumber",
    "normalize_phone_number",
    "is_phone_number",
    "is_formatted_phone_number",
    "PhoneNumber",
    "FormattedPhoneNumber",
)


class InvalidPhoneNumber(phonenumbers.NumberParseException, TypeError):
    INVALID: Final = 99

    def __init__(self, error_type: int = INVALID, msg: str = "Invalid number") -> None:
        super().__init__(error_type, msg)


def _deconstruct_phone_number(
    phone_number: str, country_code: str | None = None
) -> phonenumbers.PhoneNumber:
    try:
        parsed_number = phonenumbers.parse(phone_number, region=country_code)
    except phonenumbers.NumberParseException as e:
        raise InvalidPhoneNumber(e.error_type, e._msg)
    if not phonenumbers.is_valid_number(parsed_number):
        raise InvalidPhoneNumber
    return parsed_number


def normalize_phone_number(
    phone_number: str, country_code: str | None = None
) -> FormattedPhoneNumber:
    """
    Normalize ``phone_number`` using :py:const:`phonenumbers.PhoneNumberFormat.E164`.

    :raises InvalidPhoneNumber:
    """
    normalized = phonenumbers.format_number(
        _deconstruct_phone_number(phone_number, country_code),
        phonenumbers.PhoneNumberFormat.E164,
    )
    return cast(FormattedPhoneNumber, normalized)


is_phone_number = excepts(InvalidPhoneNumber)(_deconstruct_phone_number)
parse_str = get_bound_parser(str)


def is_formatted_phone_number(number: str) -> bool:
    try:
        return number == normalize_phone_number(number)
    except InvalidPhoneNumber:
        return False


class PhoneNumber(str, Phantom, predicate=is_phone_number):
    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A valid E.164 phone number.",
            "type": "string",
            "format": "E.164",
        }


class FormattedPhoneNumber(PhoneNumber, predicate=is_formatted_phone_number):
    @classmethod
    def parse(cls, instance: object) -> FormattedPhoneNumber:
        """
        Normalize number using :py:const:`phonenumbers.PhoneNumberFormat.E164`.

        :raises InvalidPhoneNumber:
        """
        return normalize_phone_number(parse_str(instance))

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "title": "PhoneNumber",
        }



"""
Functions that create new predicates that succeed when their argument is strictly or non
strictly between the upper and lower bounds. There are corresponding phantom types that
use these predicates in :py:mod:`phantom.interval`.
"""
from numerary.types import RealLike

from .base import Predicate
from .utils import bind_name


def open(low: RealLike, high: RealLike) -> Predicate[RealLike]:
    """
    Create a predicate that succeeds when its argument is in the range ``(low, high)``.
    """

    @bind_name(open, low, high)
    def check(value: RealLike) -> bool:
        return low <= value <= high

    return check


def open_closed(low: RealLike, high: RealLike) -> Predicate[RealLike]:
    """
    Create a predicate that succeeds when its argument is in the range ``(low, high]``.
    """

    @bind_name(open_closed, low, high)
    def check(value: RealLike) -> bool:
        return low <= value < high

    return check


def closed_open(low: RealLike, high: RealLike) -> Predicate[RealLike]:
    """
    Create a predicate that succeeds when its argument is in the range ``[low, high)``.
    """

    @bind_name(closed_open, low, high)
    def check(value: RealLike) -> bool:
        return low < value <= high

    return check


def closed(low: RealLike, high: RealLike) -> Predicate[RealLike]:
    """
    Create a predicate that succeeds when its argument is in the range ``[low, high]``.
    """

    @bind_name(closed, low, high)
    def check(value: RealLike) -> bool:
        return low < value < high

    return check



from typing import Container
from typing import Iterable
from typing import Sized
from typing import TypeVar

from .base import Predicate
from .utils import bind_name


def contains(value: object) -> Predicate[Container]:
    """Create a new predicate that succeeds when its argument contains ``value``."""

    @bind_name(contains, value)
    def compare(container: Container) -> bool:
        return value in container

    return compare


def contained(container: Container) -> Predicate[object]:
    """
    Create a new predicate that succeeds when its argument is contained by
    ``container``.
    """

    @bind_name(contained, container)
    def compare(value: object) -> bool:
        return value in container

    return compare


def count(predicate: Predicate[int]) -> Predicate[Sized]:
    """
    Create a predicate that succeeds when the size of its argument satisfies the given
    ``predicate``.
    """

    @bind_name(count, predicate)
    def compare(sized: Sized) -> bool:
        return predicate(len(sized))

    return compare


_O = TypeVar("_O", bound=object)


def exists(predicate: Predicate[_O]) -> Predicate[Iterable]:
    """
    Create a predicate that succeeds when one or more items in its argument satisfies
    ``predicate``.
    """

    @bind_name(exists, predicate)
    def compare(iterable: Iterable) -> bool:
        return any(predicate(item) for item in iterable)

    return compare


def every(predicate: Predicate[_O]) -> Predicate[Iterable]:
    """
    Create a predicate that succeeds when all items in its argument satisfy
    ``predicate``.
    """

    @bind_name(every, predicate)
    def compare(iterable: Iterable) -> bool:
        return all(predicate(item) for item in iterable)

    return compare



"""
Types for describing narrower sets of numbers than builtin numeric types like ``int``
and ``float``. Use the provided base classes to build custom intervals. For example, to
represent number in the open range ``(0, 100)`` for a volume control you would define a
type like this:

.. code-block:: python

    class VolumeLevel(int, Open, low=0, high=100):
        ...

There is also a set of concrete ready-to-use interval types provided, that use predicate
functions from :py:mod:`phantom.predicates.interval`.

.. code-block:: python

    def take_portion(portion: Portion, whole: Natural) -> float:
        return portion * whole

All interval types fully support pydantic and appropriately adds inclusive or exclusive
minimums and maximums to their schema representations.
"""

from __future__ import annotations

from typing import Any
from typing import TypeVar

from numerary.types import IntegralLike
from numerary.types import RealLike
from typing_extensions import Final
from typing_extensions import Protocol

from . import Phantom
from . import Predicate
from .predicates import interval
from .schema import Schema
from .utils import resolve_class_attr

N = TypeVar("N", bound=RealLike[IntegralLike[int]])
Derived = TypeVar("Derived", bound="Interval")


class IntervalCheck(Protocol):
    def __call__(self, a: N, b: N) -> Predicate[N]:
        ...


inf: Final = float("inf")
neg_inf: Final = float("-inf")


class Interval(Phantom[float], bound=RealLike, abstract=True):
    """
    Base class for all interval types, providing the following class arguments:

    * ``check: IntervalCheck``
    * ``low: RealLike`` (defaults to negative infinity)
    * ``high: RealLike`` (defaults to positive infinity)

    Concrete subclasses must specify their runtime type bound as their first base.
    """

    __check__: IntervalCheck
    __low__: RealLike
    __high__: RealLike

    def __init_subclass__(
        cls,
        check: IntervalCheck | None = None,
        low: RealLike = neg_inf,
        high: RealLike = inf,
        **kwargs: Any,
    ) -> None:
        resolve_class_attr(cls, "__low__", low)
        resolve_class_attr(cls, "__high__", high)
        resolve_class_attr(cls, "__check__", check)
        if getattr(cls, "__check__", None) is None:
            raise TypeError(f"{cls.__qualname__} must define an interval check")
        super().__init_subclass__(
            predicate=cls.__check__(cls.__low__, cls.__high__),
            **kwargs,
        )

    @classmethod
    def parse(cls: type[Derived], instance: object) -> Derived:
        return super().parse(
            cls.__bound__(instance) if isinstance(instance, str) else instance
        )


def _format_limit(value: RealLike) -> str:
    if value == inf:
        return "∞"
    if value == neg_inf:
        return "-∞"
    return str(value)


class Open(Interval, check=interval.open, abstract=True):
    """Uses :py:func:`phantom.predicate.interval.open` as ``check``."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": (
                f"A value in the inclusive range ({_format_limit(cls.__low__)}, "
                f"{_format_limit(cls.__high__)})."
            ),
            "minimum": cls.__low__ if cls.__low__ != neg_inf else None,
            "maximum": cls.__high__ if cls.__high__ != inf else None,
        }


class Closed(Interval, check=interval.closed, abstract=True):
    """Uses :py:func:`phantom.predicate.interval.closed` as ``check``."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": (
                f"A value in the exclusive range [{_format_limit(cls.__low__)}, "
                f"{_format_limit(cls.__high__)}]."
            ),
            "exclusiveMinimum": cls.__low__ if cls.__low__ != neg_inf else None,
            "exclusiveMaximum": cls.__high__ if cls.__high__ != inf else None,
        }


class OpenClosed(Interval, check=interval.open_closed, abstract=True):
    """Uses :py:func:`phantom.predicate.interval.open_closed` as ``check``."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": (
                f"A value in the half-open range ({_format_limit(cls.__low__)}, "
                f"{_format_limit(cls.__high__)}]."
            ),
            "minimum": cls.__low__ if cls.__low__ != neg_inf else None,
            "exclusiveMaximum": cls.__high__ if cls.__high__ != inf else None,
        }


class ClosedOpen(Interval, check=interval.closed_open, abstract=True):
    """Uses :py:func:`phantom.predicate.interval.closed_open` as ``check``."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": (
                f"A value in the half-open range [{_format_limit(cls.__low__)}, "
                f"{_format_limit(cls.__high__)})."
            ),
            "exclusiveMinimum": cls.__low__ if cls.__low__ != neg_inf else None,
            "maximum": cls.__high__ if cls.__high__ != inf else None,
        }


class Natural(int, Open, low=0):
    """Represents integer values in the inclusive range ``(0, ∞)``."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "An integer value in the inclusive range (0, ∞).",
        }


class NegativeInt(int, Open, high=0):
    """Represents integer values in the inclusive range ``(-∞, 0)``."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "An integer value in the inclusive range (-∞, 0).",
        }


class Portion(float, Open, low=0, high=1):
    """Represents float values in the inclusive range ``(0, 1)``."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A float value in the inclusive range (0, 1).",
        }



"""
Types for representing strings that match a pattern.

.. code-block:: python

    class Greeting(Match, pattern=r"^(Hi|Hello)"):
        ...


    assert isinstance("Hello Jane!", Greeting)
"""
from __future__ import annotations

import re
from typing import Any
from typing import Pattern

from .base import Phantom
from .predicates.re import is_full_match
from .predicates.re import is_match
from .schema import Schema
from .utils import resolve_class_attr

__all__ = ("Match", "FullMatch")


def _compile(pattern: Pattern[str] | str) -> Pattern[str]:
    if not isinstance(pattern, Pattern):
        return re.compile(pattern)
    return pattern


class Match(str, Phantom, abstract=True):
    """
    Takes ``pattern: Pattern[str] | str`` as class argument as either a compiled
    :py:class:`Pattern` or a :py:class:`str` to be compiled. Uses the
    :py:func:`phantom.predicate.re.is_match` predicate.
    """

    __pattern__: Pattern[str]

    def __init_subclass__(cls, pattern: Pattern[str] | str, **kwargs: Any) -> None:
        resolve_class_attr(cls, "__pattern__", _compile(pattern))
        super().__init_subclass__(predicate=is_match(cls.__pattern__), **kwargs)

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": (
                "A string starting with a match of the format regular expression."
            ),
            "format": str(cls.__pattern__.pattern),
        }


class FullMatch(str, Phantom, abstract=True):
    """
    Takes ``pattern: Pattern[str] | str`` as class argument as either a compiled
    :py:class:`Pattern` or a :py:class:`str` to be compiled. Uses the
    :py:func:`phantom.predicate.re.is_full_match` predicate.
    """

    __pattern__: Pattern[str]

    def __init_subclass__(cls, pattern: Pattern[str] | str, **kwargs: Any) -> None:
        resolve_class_attr(cls, "__pattern__", _compile(pattern))
        super().__init_subclass__(predicate=is_full_match(cls.__pattern__), **kwargs)

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A string that matches the format regular expression.",
            "format": str(cls.__pattern__.pattern),
        }



import types
from typing import Tuple
from typing import Union

import typeguard
from typing_extensions import get_args

from phantom.utils import is_union_type

from .base import Predicate
from .utils import bind_name


def equal(a: object) -> Predicate[object]:
    """Create a new predicate that succeeds when its argument is equal to ``a``."""

    @bind_name(equal, a)
    def check(b: object) -> bool:
        return a == b

    return check


def identical(a: object) -> Predicate[object]:
    """Create a new predicate that succeeds when its argument is identical to ``a``."""

    @bind_name(identical, a)
    def check(b: object) -> bool:
        return a is b

    return check


def of_type(t: Union[type, Tuple[type, ...]]) -> Predicate[object]:
    """
    Create a new predicate that succeeds when its argument is an instance of ``t``.
    """

    @bind_name(of_type, t)
    def check(a: object) -> bool:
        return isinstance(a, t)

    return check


def of_complex_type(t: type) -> Predicate[object]:
    # Hack to support PEP 604 before it's implemented in typeguard, tracking issue for
    # that feature is here: https://github.com/agronholm/typeguard/issues/222
    # Rewrite types.UnionType objects generated by PEP 604 unions as types.GenericAlias,
    # as if it was generated by a Union[].
    t = (
        types.GenericAlias(Union, get_args(t))  # type: ignore
        if is_union_type(t)
        else t
    )

    @bind_name(of_complex_type, t)
    def check(a: object) -> bool:
        try:
            typeguard.check_type("a", a, t, globals={}, locals={})
        except TypeError:
            return False
        return True

    return check



from typing import Iterable

from typing_extensions import Literal

from .base import Predicate
from .base import T
from .utils import bind_name


def true(_value: object) -> Literal[True]:
    """Always return :py:const:`True`."""
    return True


def false(_value: object) -> Literal[False]:
    """Always return :py:const:`False`."""
    return False


def negate(predicate: Predicate[T]) -> Predicate[T]:
    """Negate a given predicate."""

    @bind_name(negate, predicate)
    def check(value: T) -> bool:
        return not predicate(value)

    return check


def truthy(value: object) -> bool:
    """Return :py:const:`True` for truthy objects."""
    return bool(value)


def falsy(value: object) -> bool:
    """Return :py:const:`True` for falsy objects."""
    return negate(truthy)(value)


def both(p: Predicate[T], q: Predicate[T]) -> Predicate[T]:
    """
    Create a new predicate that succeeds when both of the given predicates succeed.
    """

    @bind_name(both, p, q)
    def check(value: T) -> bool:
        return p(value) and q(value)

    return check


def either(p: Predicate[T], q: Predicate[T]) -> Predicate[T]:
    """
    Create a new predicate that succeeds when at least one of the given predicates
    succeed.
    """

    @bind_name(either, p, q)
    def check(value: T) -> bool:
        return p(value) or q(value)

    return check


def xor(p: Predicate[T], q: Predicate[T]) -> Predicate[T]:
    """
    Create a new predicate that succeeds when one of the given predicates succeed, but
    not both.
    """

    @bind_name(xor, p, q)
    def check(value: T) -> bool:
        return p(value) ^ q(value)

    return check


def all_of(predicates: Iterable[Predicate[T]]) -> Predicate[T]:
    """Create a new predicate that succeeds when all of the given predicates succeed."""
    predicates = tuple(predicates)

    @bind_name(all_of, *predicates)
    def check(value: T) -> bool:
        return all(p(value) for p in predicates)

    return check


def any_of(predicates: Iterable[Predicate[T]]) -> Predicate[T]:
    """
    Create a new predicate that succeeds when at least one of the given predicates
    succeed.
    """
    predicates = tuple(predicates)

    @bind_name(any_of, *predicates)
    def check(value: T) -> bool:
        return any(p(value) for p in predicates)

    return check


def one_of(predicates: Iterable[Predicate[T]]) -> Predicate[T]:
    """
    Create a new predicate that succeeds when exactly one of the given predicates
    succeed.
    """
    predicates = tuple(predicates)

    @bind_name(one_of, *predicates)
    def check(value: T) -> bool:
        return 1 == sum(p(value) for p in predicates)

    return check



"""
Types describing objects that coerce to either ``True`` or ``False`` respectively when
calling ``bool()`` on them.
"""

from .base import Phantom
from .predicates import boolean


class Truthy(Phantom[object], predicate=boolean.truthy, bound=object):
    """
    >>> isinstance("Huzzah!", Truthy)
    True
    >>> isinstance((), Truthy)
    False
    """


class Falsy(Phantom[object], predicate=boolean.falsy, bound=object):
    """
    >>> isinstance((), Falsy)
    True
    >>> isinstance("Hej!", Falsy)
    False
    """



from typing import Pattern

from .base import Predicate
from .utils import bind_name


def is_match(pattern: Pattern[str]) -> Predicate[str]:
    """
    Create a new predicate that succeeds when the start of its argument matches the
    given ``pattern``.
    """

    @bind_name(is_match, pattern.pattern)
    def match(instance: str) -> bool:
        return pattern.match(instance) is not None

    return match


def is_full_match(pattern: Pattern[str]) -> Predicate[str]:
    """
    Create a new predicate that succeeds when its whole argument matches the given
    ``pattern``.
    """

    @bind_name(is_full_match, pattern.pattern)
    def full_match(instance: str) -> bool:
        return pattern.fullmatch(instance) is not None

    return full_match



from __future__ import annotations

import functools
from functools import partial
from typing import Any
from typing import Callable
from typing import TypeVar


def _name(fn: Callable) -> str:
    if isinstance(fn, partial):
        fn = fn.func
    try:
        return fn.__qualname__
    except AttributeError:
        return str(fn)


AA = TypeVar("AA")
AR = TypeVar("AR")
BA = TypeVar("BA")


def compose2(a: Callable[[AA], AR], b: Callable[[BA], AA]) -> Callable[[BA], AR]:
    """
    Returns a function composed from the two given functions ``a`` and ``b`` such that
    calling ``compose2(a, b)(x)`` is equivalent to calling ``a(b(x))``.

    >>> compose2("".join, reversed)("!olleH")
    'Hello!'
    """
    a_name = _name(a)
    b_name = _name(b)

    def c(arg: BA) -> AR:
        return a(b(arg))

    c.__name__ = f"{a_name}∘{b_name}"
    c.__doc__ = f"Function composed as {a_name}({b_name}(_))."
    return c


A = TypeVar("A")


def excepts(
    exception: tuple[type[Exception], ...] | type[Exception],
    negate: bool = False,
) -> Callable[[Callable[[A], Any]], Callable[[A], bool]]:
    """
    Turn a unary function that raises an exception into a boolean predicate.

    >>> def validate_positive(number: int) -> None:
    ...     if number < 0: raise ValueError
    >>> is_positive = excepts(ValueError)(validate_positive)
    >>> is_positive(0), is_positive(-1)
    (True, False)
    """

    def decorator(fn: Callable[[A], Any]) -> Callable[[A], bool]:
        @functools.wraps(fn)
        def wrapper(arg: A) -> bool:
            try:
                fn(arg)
            except exception:
                return negate
            return not negate

        return wrapper

    return decorator



from __future__ import annotations

import abc
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import Sequence
from typing import TypeVar
from typing import cast

from typing_extensions import Protocol
from typing_extensions import get_args
from typing_extensions import runtime_checkable

from .predicates.base import Predicate
from .predicates.boolean import all_of
from .predicates.generic import of_complex_type
from .predicates.generic import of_type
from .schema import SchemaField
from .utils import BoundType
from .utils import NotKnownMutable
from .utils import UnresolvedClassAttribute
from .utils import fully_qualified_name
from .utils import is_not_mutable
from .utils import is_subtype
from .utils import is_union
from .utils import resolve_class_attr


@runtime_checkable
class InstanceCheckable(Protocol):
    def __instancecheck__(self, instance: object) -> bool:
        ...


class PhantomMeta(abc.ABCMeta):
    """
    Metaclass that defers __instancecheck__ to derived classes and prevents actual
    instance creation.
    """

    def __instancecheck__(self, instance: object) -> bool:
        if not issubclass(self, InstanceCheckable):
            return False
        return self.__instancecheck__(instance)

    # With the current level of metaclass support in mypy it's unlikely that we'll be
    # able to make this context typed, hence the ignores.
    def __call__(cls, instance):  # type: ignore[no-untyped-def]
        return cls.parse(instance)  # type: ignore[attr-defined]


class BoundError(TypeError):
    ...


T = TypeVar("T", covariant=True)


def display_bound(bound: Any) -> str:
    if isinstance(bound, Iterable):
        return f"Intersection[{', '.join(display_bound(part) for part in bound)}]"
    if is_union(bound):
        return (
            f"typing.Union["
            f"{', '.join(display_bound(part) for part in get_args(bound))}"
            f"]"
        )
    return str(getattr(bound, "__name__", bound))


def get_bound_parser(bound: Any) -> Callable[[object], T]:
    within_bound = (
        # Interpret sequence as intersection
        all_of(of_type(t) for t in bound)
        if isinstance(bound, Sequence)
        else of_complex_type(bound)
    )

    def parser(instance: object) -> T:
        if not within_bound(instance):
            raise BoundError(
                f"Value is not within bound of {display_bound(bound)!r}: {instance!r}"
            )
        return cast(T, instance)

    return parser


Derived = TypeVar("Derived", bound="PhantomBase")


class PhantomBase(SchemaField, metaclass=PhantomMeta):
    @classmethod
    def parse(cls: type[Derived], instance: object) -> Derived:
        """
        Parse an arbitrary value into a phantom type.

        :raises TypeError:
        """
        if not isinstance(instance, cls):
            raise TypeError(
                f"Could not parse {fully_qualified_name(cls)} from {instance!r}"
            )
        return instance

    @classmethod
    @abc.abstractmethod
    def __instancecheck__(cls, instance: object) -> bool:
        ...

    @classmethod
    def __get_validators__(cls: type[Derived]) -> Iterator[Callable[[object], Derived]]:
        """Hook that makes phantom types compatible with pydantic."""
        yield cls.parse


class AbstractInstanceCheck(TypeError):
    ...


class Phantom(PhantomBase, Generic[T]):
    """
    Base class for predicate-based phantom types.

    **Class arguments**

    * ``predicate: Predicate[T] | None`` - Predicate function used for instance checks.
      Can be ``None`` if the type is abstract.
    * ``bound: type[T] | None`` - Bound used to check values before passing them to the
      type's predicate function. This will often but not always be the same as the
      runtime type that values of the phantom type are represented as. If this is not
      provided as a class argument, it's attempted to be resolved in order from an
      implicit bound (any bases of the type that come before ``Phantom``), or inherited
      from super phantom types that provide a bound. Can be ``None`` if the type is
      abstract.
    * ``abstract: bool`` - Set to ``True`` to create an abstract phantom type. This
      allows deferring definitions of ``predicate`` and ``bound`` to concrete subtypes.
    """

    __predicate__: Predicate[T]
    # The bound of a phantom type is the type that its values will have at
    # runtime, so when checking if a value is an instance of a phantom type,
    # it's first checked to be within its bounds, so that the value can be
    # safely passed as argument to the predicate function.
    #
    # When subclassing, the bound of the new type must be a subtype of the bound
    # of the super class.
    __bound__: ClassVar[NotKnownMutable]
    __abstract__: ClassVar[bool]

    def __init_subclass__(
        cls,
        predicate: Predicate[T] | None = None,
        bound: type[T] | None = None,
        abstract: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)
        resolve_class_attr(cls, "__abstract__", abstract)
        resolve_class_attr(cls, "__predicate__", predicate)
        cls._resolve_bound(bound)

    @classmethod
    def _interpret_implicit_bound(cls) -> BoundType:
        def discover_bounds() -> Iterable[type]:
            for type_ in cls.__mro__:
                if type_ is cls:
                    continue
                if issubclass(type_, Phantom):
                    break
                yield type_
            else:  # pragma: no cover
                raise RuntimeError(f"{cls} is not a subclass of Phantom")

        types = tuple(discover_bounds())
        if len(types) == 1:
            return types[0]
        return types

    @classmethod
    def _resolve_bound(cls, class_arg: Any) -> None:
        inherited = getattr(cls, "__bound__", None)
        implicit = cls._interpret_implicit_bound()
        if class_arg is not None:
            bound = class_arg
        elif implicit:
            bound = implicit
        elif inherited is not None:
            bound = inherited
        elif not getattr(cls, "__abstract__", False):
            raise UnresolvedClassAttribute(
                f"Concrete phantom type {cls.__qualname__} must define class attribute "
                f"__bound__."
            )
        else:
            return

        if inherited is not None and not is_subtype(bound, inherited):
            raise BoundError(
                f"The bounds of {cls.__qualname__} are not compatible with its "
                f"inherited bounds."
            )

        assert is_not_mutable(bound)

        cls.__bound__ = bound

    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        if cls.__abstract__:
            raise AbstractInstanceCheck(
                "Abstract phantom types cannot be used in instance checks"
            )
        try:
            instance = get_bound_parser(cls.__bound__)(instance)
        except BoundError:
            return False
        return cls.__predicate__(instance)



from numerary.types import RealLike

from .base import Predicate
from .boolean import negate
from .generic import equal
from .utils import bind_name


def less(n: RealLike) -> Predicate[RealLike]:
    """
    Create a new predicate that succeeds when its argument is strictly less than ``n``.
    """

    @bind_name(less, n)
    def check(value: RealLike) -> bool:
        return value < n

    return check


def le(n: RealLike) -> Predicate[RealLike]:
    """
    Create a new predicate that succeeds when its argument is less than or equal to
    ``n``.
    """

    @bind_name(le, n)
    def check(value: RealLike) -> bool:
        return value <= n

    return check


def greater(n: RealLike) -> Predicate[RealLike]:
    """
    Create a new predicate that succeeds when its argument is strictly greater than
    ``n``.
    """

    @bind_name(greater, n)
    def check(value: RealLike) -> bool:
        return value > n

    return check


def ge(n: RealLike) -> Predicate[RealLike]:
    """
    Create a new predicate that succeeds when its argument is greater than or equal to
    ``n``.
    """

    @bind_name(ge, n)
    def check(value: RealLike) -> bool:
        return value >= n

    return check


def positive(n: RealLike) -> bool:
    """Return :py:const:`True` when ``n`` is strictly greater than zero."""
    return greater(0)(n)


def non_positive(n: RealLike) -> bool:
    """Return :py:const:`True` when ``n``  is less than or equal to zero."""
    return le(0)(n)


def negative(n: RealLike) -> bool:
    """Return :py:const:`True` when ``n`` is strictly less than zero."""
    return less(0)(n)


def non_negative(n: RealLike) -> bool:
    """Return :py:const:`True` when ``n`` is greater than or equal to zero."""
    return ge(0)(n)


def modulo(n: RealLike, p: Predicate[RealLike]) -> Predicate[RealLike]:
    """
    Create a new predicate that succeeds when its argument modulo ``n`` satisfies the
    given predicate ``p``.
    """

    @bind_name(modulo, n, p)
    def check(value: RealLike) -> bool:
        return p(value % n)

    return check


def even(n: int) -> bool:
    """Return :py:const:`True`  when ``n`` is even."""
    return modulo(2, equal(0))(n)


def odd(n: int) -> bool:
    """Return :py:const:`True`  when ``n`` is odd."""
    return negate(even)(n)



from typing import Callable
from typing import TypeVar


def _name_or_repr(obj: object) -> str:
    try:
        return str(obj.__qualname__)  # type: ignore[attr-defined]
    except AttributeError:
        return repr(obj)


B = TypeVar("B", bound=Callable)


def bind_name(wrapped: Callable, *values: object) -> Callable[[B], B]:
    name = (
        f"{wrapped.__qualname__}({', '.join(_name_or_repr(value) for value in values)})"
    )

    def decorator(inner: B) -> B:
        inner.__qualname__ = inner.__name__ = name
        return inner

    return decorator



"""
Types describing collections with size boundaries. These types should only be used with
immutable collections. There is a naive check that eliminates some of the most common
mutable collections in the instance check. However, a guaranteed check is probably
impossible to implement, so some amount of developer discipline is required.

Sized types are created by subclassing ``PhantomSized`` and providing a predicate that
will be called with the size of the tested collection. For instance, ``NonEmpty`` is
implemented using ``len=numeric.greater(0)``.

This made-up type would describe sized collections with between 5 and 10 ints:

.. code-block:: python

    class SpecificSize(PhantomSized[int], len=interval.open(5, 10)):
        ...
"""
from typing import Any
from typing import Generic
from typing import Iterable
from typing import MutableMapping
from typing import MutableSequence
from typing import MutableSet
from typing import Sized
from typing import TypeVar

from numerary.types import RealLike

# We attempt to import _ProtocolMeta from typing_extensions to support Python 3.7 but
# fall back the typing module to support Python 3.8+. This is the closest I could find
# to documentation of _ProtocolMeta.
# https://github.com/python/cpython/commit/74d7f76e2c953fbfdb7ce01b7319d91d471cc5ef
try:
    from typing_extensions import _ProtocolMeta  # type: ignore[attr-defined]
except ImportError:
    from typing import _ProtocolMeta  # type: ignore[attr-defined]

from typing_extensions import Final
from typing_extensions import Protocol
from typing_extensions import runtime_checkable

from . import Phantom
from . import PhantomMeta
from . import Predicate
from .predicates import boolean
from .predicates import collection
from .predicates import generic
from .predicates import numeric
from .schema import Schema

__all__ = (
    "SizedIterable",
    "PhantomSized",
    "NonEmpty",
    "Empty",
)


mutable: Final = (MutableSequence, MutableSet, MutableMapping)
T = TypeVar("T", bound=object, covariant=True)


@runtime_checkable
class SizedIterable(Sized, Iterable[T], Protocol[T]):
    """Intersection of :py:class:`typing.Sized` and :py:class:`typing.Iterable`."""


# This raises a mypy error because disallow_subclassing_any is enabled and _ProtocolMeta
# isn't publicly typed.
class SizedIterablePhantomMeta(PhantomMeta, _ProtocolMeta):  # type: ignore[misc]
    ...


class PhantomSized(
    Phantom[Sized],
    SizedIterable[T],
    Generic[T],
    metaclass=SizedIterablePhantomMeta,
    bound=SizedIterable,
    abstract=True,
):
    """Takes class argument ``len: Predicate[RealLike]``."""

    def __init_subclass__(cls, len: Predicate[RealLike], **kwargs: Any) -> None:
        super().__init_subclass__(
            predicate=boolean.both(
                boolean.negate(generic.of_type(mutable)),
                collection.count(len),
            ),
            **kwargs,
        )

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "type": "array",
        }


class NonEmpty(PhantomSized[T], Generic[T], len=numeric.greater(0)):
    """A sized collection with at least one item."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "A non-empty array.",
            "minItems": 1,
        }


class Empty(PhantomSized[T], Generic[T], len=generic.equal(0)):
    """A sized collection with exactly zero items."""

    @classmethod
    def __schema__(cls) -> Schema:
        return {
            **super().__schema__(),  # type: ignore[misc]
            "description": "An empty array.",
            "maxItems": 0,
        }
